@@ -43,7 +43,7 @@
         ]).
 
 -export([
-         always_ok/1,
+         always_ok/0,
          app_test/1,
          appup_test/1,
          basic_test/1,
@@ -109,7 +109,9 @@
 %%--------------------------------------------------------------------
 
 suite() ->
-    [{ct_hooks,[ts_install_cth]},
+    [{ct_hooks,[ts_install_cth,
+                {cth_events,
+                 [{verify_fun, fun verify_events/2}]}]},
      {timetrap,{seconds,90}}].
 
 all() -> 
@@ -190,8 +192,9 @@ init_per_group(_, Config) ->
 end_per_group(_, Config) ->
     Config.
 %%--------------------------------------------------------------------
-init_per_testcase(TC, Config) when TC==shell_no_unicode ; 
-				   TC==shell_unicode_string ->
+init_per_testcase(TestCase, Config)
+  when TestCase==shell_no_unicode;
+       TestCase==shell_unicode_string ->
     PrivDir = proplists:get_value(priv_dir, Config),
     UserDir = proplists:get_value(priv_dir, Config),
     SysDir =  proplists:get_value(data_dir, Config),
@@ -208,7 +211,6 @@ init_per_testcase(TC, Config) when TC==shell_no_unicode ;
     ct:log("file:native_name_encoding() = ~p,~nio:getopts() = ~p",
 	   [file:native_name_encoding(),io:getopts()]),
     wait_for_erlang_first_line([{io,IO}, {shell,Shell}, {sftpd, Sftpd}  | Config]);
-
 init_per_testcase(inet6_option, Config) ->
     case ssh_test_lib:has_inet6_address() of
 	true ->
@@ -216,24 +218,31 @@ init_per_testcase(inet6_option, Config) ->
 	false ->
 	    {skip,"No ipv6 interface address"}
     end;
-init_per_testcase(_TestCase, Config) ->
-    Config.
+init_per_testcase(TestCase, Config) ->
+    ssh_test_lib:add_log_handler(TestCase, Config).
 
-end_per_testcase(TC, Config) when TC==shell_no_unicode ; 
-				  TC==shell_unicode_string ->
+end_per_testcase(TestCase, Config)
+  when TestCase==shell_no_unicode;
+       TestCase==shell_unicode_string ->
     case proplists:get_value(sftpd, Config) of
 	{Pid, _, _} ->
-	    catch ssh:stop_daemon(Pid);
+	    catch ssh:stop_daemon(Pid),
+            ok;
 	_ ->
 	    ok
-    end,
-    end_per_testcase(Config);
-end_per_testcase(_TestCase, Config) ->
-    end_per_testcase(Config).
-
-end_per_testcase(_Config) ->
+    end;
+end_per_testcase(_TestCase, _Config) ->
     ok.
 
+verify_events(_TestCase, 0) ->
+    ok;
+verify_events(multi_daemon_opt_fd, 6) -> ok;
+verify_events(internal_error, 3) -> ok;
+verify_events(_TestCase, EventNumber) when EventNumber > 0->
+    {fail, lists:flatten(
+             io_lib:format("unexpected event cnt: ~s",
+                           [integer_to_list(EventNumber)]))}.
+
 %%--------------------------------------------------------------------
 %% Test Cases --------------------------------------------------------
 %%--------------------------------------------------------------------
@@ -1051,7 +1060,7 @@ parallel_login(Config) when is_list(Config) ->
     ok = ssh_connection:send(ConnectionRef, ChannelId, <<"Data">>),
     ok = ssh_connection:send(ConnectionRef, ChannelId, << >>),
     ssh_info:print(fun(Fmt, Args) -> io:fwrite(user, Fmt, Args) end),
-    {Parents, Conns, Handshakers} =
+    {_Parents, _Conns, _Handshakers} =
         ssh_test_lib:find_handshake_parent(Port),
     ssh:stop_daemon(Pid).
 
@@ -1465,7 +1474,7 @@ shell_exit_status(Config) when is_list(Config) ->
     ssh_test_lib:receive_exec_end(ConnectionRef, ChannelId),
     ssh:stop_daemon(Pid).
 
-always_ok(_) -> ok.
+always_ok() -> spawn(fun() -> exit(ok) end).
     
 %%----------------------------------------------------------------------------
 setopts_getopts(Config) ->
@@ -1539,7 +1548,7 @@ wait_for_erlang_first_line(Config) ->
 	    {fail,no_ssh_connection};
 	<<"Eshell ",_/binary>> = _ErlShellStart ->
 	    ct:log("Erlang shell start: ~p~n", [_ErlShellStart]),
-	    Config;
+            Config;
 	Other ->
 	    ct:log("Unexpected answer from ssh server: ~p",[Other]),
 	    {fail,unexpected_answer}
