@@ -1,7 +1,9 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1997-2023. All Rights Reserved.
+%% SPDX-License-Identifier: Apache-2.0
+%%
+%% Copyright Ericsson AB 1997-2025. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -18,6 +20,38 @@
 %% %CopyrightEnd%
 
 -module(filelib).
+-moduledoc """
+File utilities, such as wildcard matching of filenames.
+
+This module contains utilities on a higher level than the `m:file` module.
+
+This module does not support "raw" filenames (that is, files whose names do not
+comply with the expected encoding). Such files are ignored by the functions in
+this module.
+
+For more information about raw filenames, see the `m:file` module.
+
+> #### Note {: .info }
+>
+> Functionality in this module generally assumes valid input and does not
+> necessarily fail on input that does not use a valid encoding, but may instead
+> very likely produce invalid output.
+>
+> File operations used to accept filenames containing null characters (integer
+> value zero). This caused the name to be truncated and in some cases arguments
+> to primitive operations to be mixed up. Filenames containing null characters
+> inside the filename are now _rejected_ and will cause primitive file
+> operations to fail.
+
+> #### Warning {: .warning }
+>
+> Currently null characters at the end of the filename will be accepted by
+> primitive file operations. Such filenames are however still documented as
+> invalid. The implementation will also change in the future and reject such
+> filenames.
+""".
+
+-compile(nowarn_deprecated_catch).
 
 %% File utilities.
 -export([wildcard/1, wildcard/2, is_dir/1, is_file/1, is_regular/1]).
@@ -49,11 +83,88 @@
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
+-doc """
+Returns a list of all files that match Unix-style wildcard string `Wildcard`.
+
+The wildcard string looks like an ordinary filename, except that the following
+"wildcard characters" are interpreted in a special way:
+
+- **?** - Matches one character.
+
+- **\*** - Matches any number of characters up to the end of the filename, the
+  next dot, or the next slash.
+
+- **\*\*** - Two adjacent `*` used as a single pattern match all files and zero
+  or more directories and subdirectories.
+
+- **\[Character1,Character2,...]** - Matches any of the characters listed. Two
+  characters separated by a hyphen match a range of characters. Example: `[A-Z]`
+  matches any uppercase letter.
+
+- **\{Item,...\}** - Alternation. Matches one of the alternatives.
+
+Other characters represent themselves. Only filenames that have exactly the same
+character in the same position match. Matching is case-sensitive, for example,
+"a" does not match "A".
+
+Directory separators must always be written as `/`, even on Windows.
+
+A character preceded by `\` loses its special meaning. Note that `\` must be
+written as `\\` in a string literal. For example, "\\\\?\*" will match any
+filename starting with `?`.
+
+Notice that multiple "\*" characters are allowed (as in Unix wildcards, but
+opposed to Windows/DOS wildcards).
+
+_Examples:_
+
+The following examples assume that the current directory is the top of an
+Erlang/OTP installation.
+
+To find all `.beam` files in all applications, use the following line:
+
+```text
+filelib:wildcard("lib/*/ebin/*.beam").
+```
+
+To find `.erl` or `.hrl` in all applications `src` directories, use either of
+the following lines:
+
+```text
+filelib:wildcard("lib/*/src/*.?rl")
+```
+
+```text
+filelib:wildcard("lib/*/src/*.{erl,hrl}")
+```
+
+To find all `.hrl` files in `src` or `include` directories:
+
+```text
+filelib:wildcard("lib/*/{src,include}/*.hrl").
+```
+
+To find all `.erl` or `.hrl` files in either `src` or `include` directories:
+
+```text
+filelib:wildcard("lib/*/{src,include}/*.{erl,hrl}")
+```
+
+To find all `.erl` or `.hrl` files in any subdirectory:
+
+```text
+filelib:wildcard("lib/**/*.{erl,hrl}")
+```
+""".
 -spec wildcard(Wildcard) -> [file:filename()] when
       Wildcard :: filename() | dirname().
 wildcard(Pattern) when is_list(Pattern) ->
     ?HANDLE_ERROR(do_wildcard(Pattern, ".", file)).
 
+-doc """
+Same as `wildcard/1`, except that `Cwd` is used instead of the working
+directory.
+""".
 -spec wildcard(Wildcard, Cwd) -> [file:filename()] when
       Wildcard :: filename() | dirname(),
       Cwd :: dirname().
@@ -62,38 +173,63 @@ wildcard(Pattern, Cwd) when is_list(Pattern), is_list(Cwd) ->
 wildcard(Pattern, Mod) when is_list(Pattern), is_atom(Mod) ->
     ?HANDLE_ERROR(do_wildcard(Pattern, ".", Mod)).
 
+-doc false.
 -spec wildcard(file:name(), file:name(), atom()) -> [file:filename()].
 wildcard(Pattern, Cwd, Mod)
   when is_list(Pattern), is_list(Cwd), is_atom(Mod) ->
     ?HANDLE_ERROR(do_wildcard(Pattern, Cwd, Mod)).
 
+-doc "Returns `true` if `Name` refers to a directory, otherwise `false`.".
 -spec is_dir(Name) -> boolean() when
       Name :: filename_all() | dirname_all().
 is_dir(Dir) ->
     do_is_dir(Dir, file).
 
+-doc false.
 -spec is_dir(file:name_all(), atom()) -> boolean().
 is_dir(Dir, Mod) when is_atom(Mod) ->
     do_is_dir(Dir, Mod).
 
+-doc "Returns `true` if `Name` refers to a file or a directory, otherwise `false`.".
 -spec is_file(Name) -> boolean() when
       Name :: filename_all() | dirname_all().
 is_file(File) ->
     do_is_file(File, file).
 
+-doc false.
 -spec is_file(file:name_all(), atom()) -> boolean().
 is_file(File, Mod) when is_atom(Mod) ->
     do_is_file(File, Mod).
 
+-doc "Returns `true` if `Name` refers to a (regular) file, otherwise `false`.".
 -spec is_regular(Name) -> boolean() when
       Name :: filename_all().
 is_regular(File) ->
     do_is_regular(File, file).
     
+-doc false.
 -spec is_regular(file:name_all(), atom()) -> boolean().
 is_regular(File, Mod) when is_atom(Mod) ->
     do_is_regular(File, Mod).
     
+-doc """
+Folds function `Fun` over all (regular) files `F` in directory `Dir` whose
+basename (for example, just `"baz.erl"` in `"foo/bar/baz.erl"`) matches the
+regular expression `RegExp` (for a description of the allowed regular
+expressions, see the `m:re` module).
+
+If `Recursive` is `true`, all subdirectories to `Dir` are processed.
+The regular expression matching is only done on the filename without the directory part.
+
+If Unicode filename translation is in effect and the file system is transparent,
+filenames that cannot be interpreted as Unicode can be encountered, in which
+case the `fun()` must be prepared to handle raw filenames (that is, binaries).
+If the regular expression contains codepoints > 255, it does not match filenames
+that do not conform to the expected character encoding (that is, are not encoded
+in valid UTF-8).
+
+For more information about raw filenames, see the `m:file` module.
+""".
 -spec fold_files(Dir, RegExp, Recursive, Fun, AccIn) -> AccOut when
       Dir :: dirname(),
       RegExp :: string(),
@@ -104,24 +240,32 @@ is_regular(File, Mod) when is_atom(Mod) ->
 fold_files(Dir, RegExp, Recursive, Fun, Acc) ->
     do_fold_files(Dir, RegExp, Recursive, Fun, Acc, file).
 
+-doc false.
 -spec fold_files(file:name(), string(), boolean(), fun((_,_) -> _), _, atom()) -> _.
 fold_files(Dir, RegExp, Recursive, Fun, Acc, Mod) when is_atom(Mod) ->
     do_fold_files(Dir, RegExp, Recursive, Fun, Acc, Mod).
 
+-doc """
+Returns the date and time the specified file or directory was last modified, or
+`0` if the file does not exist.
+""".
 -spec last_modified(Name) -> file:date_time() | 0 when
       Name :: filename_all() | dirname_all().
 last_modified(File) ->
     do_last_modified(File, file).
 
+-doc false.
 -spec last_modified(file:name_all(), atom()) -> file:date_time() | 0.
 last_modified(File, Mod) when is_atom(Mod) ->
     do_last_modified(File, Mod).
 
+-doc "Returns the size of the specified file.".
 -spec file_size(Filename) -> non_neg_integer() when
       Filename :: filename_all().
 file_size(File) ->
     do_file_size(File, file).
 
+-doc false.
 -spec file_size(file:name(), atom()) -> non_neg_integer().
 file_size(File, Mod) when is_atom(Mod) ->
     do_file_size(File, Mod).
@@ -224,6 +368,13 @@ do_file_size(File, Mod) ->
 %% +type X = filename() | dirname()
 %% ensures that the directory name required to create D exists
 
+-doc """
+Ensures that all parent directories for the specified file or directory name
+`Name` exist, trying to create them if necessary.
+
+Returns `ok` if all parent directories already exist or can be created. Returns
+`{error, Reason}` if some parent directory does not exist and cannot be created.
+""".
 -spec ensure_dir(Name) -> 'ok' | {'error', Reason} when
       Name :: filename_all() | dirname_all(),
       Reason :: file:posix().
@@ -233,6 +384,17 @@ ensure_dir(F) ->
     Dir = filename:dirname(F),
     ensure_path(Dir).
 
+-doc """
+Ensures that all parent directories for the specified path `Path` exist, trying
+to create them if necessary.
+
+Unlike `ensure_dir/1`, this function will attempt to create all path segments as
+a directory, including the last segment.
+
+Returns `ok` if all parent directories already exist or can be created. Returns
+`{error, Reason}` if some parent directory does not exist and cannot be created.
+""".
+-doc(#{since => <<"OTP 25.0">>}).
 -spec ensure_path(Path) -> 'ok' | {'error', Reason} when
       Path :: dirname_all(),
       Reason :: file:posix().
@@ -404,6 +566,7 @@ do_list_dir(Dir, Mod) ->     eval_list_dir(Dir, Mod).
 -define(ESCAPED_ESCAPE_PREFIX, [?ESCAPE_PREFIX,?ESCAPE_PREFIX]).
 
 %% Only for debugging.
+-doc false.
 compile_wildcard(Pattern) when is_list(Pattern) ->
     {compiled_wildcard,?HANDLE_ERROR(compile_wildcard(Pattern, "."))}.
 
@@ -680,11 +843,25 @@ asn1_source_search_rules() ->
 
 -type find_file_rule() :: {ObjDirSuffix::string(), SrcDirSuffix::string()}.
 
--spec find_file(filename(), filename()) ->
+-doc(#{equiv => find_file(Filename, Dir, [])}).
+-doc(#{since => <<"OTP 20.0">>}).
+-spec find_file(Filename :: filename(), Dir :: filename()) ->
         {ok, filename()} | {error, not_found}.
 find_file(Filename, Dir) ->
     find_file(Filename, Dir, []).
 
+-doc """
+Looks for a file of the given name by applying suffix rules to the given
+directory path.
+
+For example, a rule `{"ebin", "src"}` means that if the directory path ends with
+ `"ebin"`, the corresponding path ending in `"src"` should be searched.
+
+If `Rules` is left out or is an empty list, the default system rules are used.
+See also the Kernel application parameter
+[`source_search_rules`](`e:kernel:kernel_app.md#source_search_rules`).
+""".
+-doc(#{since => <<"OTP 20.0">>}).
 -spec find_file(filename(), filename(), [find_file_rule()]) ->
         {ok, filename()} | {error, not_found}.
 find_file(Filename, Dir, []) ->
@@ -697,16 +874,38 @@ find_file(Filename, Dir, Rules) ->
 -type find_source_rule() :: {ObjExtension::string(), SrcExtension::string(),
                              [find_file_rule()]}.
 
+-doc """
+Equivalent to [`find_source(Base, Dir)`](`find_source/2`), where `Dir` is
+`filename:dirname(FilePath)` and `Base` is `filename:basename(FilePath)`.
+""".
+-doc(#{since => <<"OTP 20.0">>}).
 -spec find_source(filename()) ->
         {ok, filename()} | {error, not_found}.
 find_source(FilePath) ->
     find_source(filename:basename(FilePath), filename:dirname(FilePath)).
 
+-doc(#{equiv => find_source(Filename, Dir, [])}).
+-doc(#{since => <<"OTP 20.0">>}).
 -spec find_source(filename(), filename()) ->
         {ok, filename()} | {error, not_found}.
 find_source(Filename, Dir) ->
     find_source(Filename, Dir, []).
 
+-doc """
+Applies file extension specific rules to find the source file for a given object
+file relative to the object directory.
+
+For example, for a file with the extension `.beam`, the default rule is to look
+for a file with a corresponding extension `.erl` by replacing the suffix `"ebin"`
+of the object directory path with `"src"` or `"src/*"`. The file search is done
+through `find_file/3`. The directory of the object file is always tried before
+any other directory specified by the rules.
+
+If `Rules` is left out or is an empty list, the default system rules are used.
+See also the Kernel application parameter
+[`source_search_rules`](`e:kernel:kernel_app.md#source_search_rules`).
+""".
+-doc(#{since => <<"OTP 20.0">>}).
 -spec find_source(filename(), filename(), [find_source_rule()]) ->
         {ok, filename()} | {error, not_found}.
 find_source(Filename, Dir, []) ->
@@ -765,6 +964,33 @@ find_regular_file([File|Files]) ->
         false -> find_regular_file(Files)
     end.
 
+-doc """
+Sanitizes the relative path by eliminating ".." and "." components to protect
+against directory traversal attacks.
+
+Either returns the sanitized path name, or the atom `unsafe` if the path is unsafe.
+The path is considered unsafe in the following circumstances:
+
+- The path is not relative.
+- A ".." component would climb up above the root of the relative path.
+- A symbolic link in the path points above the root of the relative path.
+
+_Examples:_
+
+```erlang
+1> {ok, Cwd} = file:get_cwd().
+...
+2> filelib:safe_relative_path("dir/sub_dir/..", Cwd).
+"dir"
+3> filelib:safe_relative_path("dir/..", Cwd).
+[]
+4> filelib:safe_relative_path("dir/../..", Cwd).
+unsafe
+5> filelib:safe_relative_path("/abs/path", Cwd).
+unsafe
+```
+""".
+-doc(#{since => <<"OTP 23.0">>}).
 -spec safe_relative_path(Filename, Cwd) -> unsafe | SafeFilename when
       Filename :: filename_all(),
       Cwd :: filename_all(),
@@ -775,7 +1001,7 @@ safe_relative_path(Path, "") ->
 safe_relative_path(Path, Cwd) ->
     srp_path(filename:split(Path),
              Cwd,
-             sets:new([{version, 2}]),
+             sets:new(),
              []).
 
 srp_path([], _Cwd, _Seen, []) ->
@@ -795,7 +1021,7 @@ srp_path([<<"..">>|_Segs], _Cwd, _Seen, []) ->
 srp_path([<<"..">>|Segs], Cwd, Seen, [_|_]=Acc) ->
     srp_path(Segs, Cwd, Seen, lists:droplast(Acc));
 srp_path([clear|Segs], Cwd, _Seen, Acc) ->
-    srp_path(Segs, Cwd, sets:new([{version, 2}]), Acc);
+    srp_path(Segs, Cwd, sets:new(), Acc);
 srp_path([Seg|_]=Segs, Cwd, Seen, Acc) ->
     case filename:pathtype(Seg) of
         relative ->
